package zero1hd.rhythmbullet.desktop.audio.visualizer;

import static org.lwjgl.openal.AL10.alGetSourcef;

import java.nio.ByteBuffer;
import java.nio.ShortBuffer;

import org.lwjgl.openal.AL11;

import com.badlogic.gdx.Gdx;
import com.badlogic.gdx.backends.lwjgl.audio.OpenALMusic;
import com.badlogic.gdx.graphics.g2d.Batch;
import com.badlogic.gdx.utils.reflect.ClassReflection;
import com.badlogic.gdx.utils.reflect.Field;
import com.badlogic.gdx.utils.reflect.ReflectionException;

import zero1hd.rhythmbullet.audio.MusicManager;
import zero1hd.rhythmbullet.audio.visualizer.MusicManagerFFT;
import zero1hd.rhythmbullet.audio.visualizer.Visualizer;

public class DesktopVisualizer extends MusicManagerFFT implements Visualizer {
	private ShortBuffer playingBuffer;
	private ShortBuffer compareBuffer;
	private ShortBuffer buffer;
	private int sourceID;
	private int readWindowIndex;

	public DesktopVisualizer() {
		super();
		try {
			Field bufferField = ClassReflection.getDeclaredField(OpenALMusic.class, "tempBuffer");
			bufferField.setAccessible(true);
			buffer = ((ByteBuffer) bufferField.get(null)).asShortBuffer().asReadOnlyBuffer();
		} catch (IllegalArgumentException | SecurityException | ReflectionException e) {
			e.printStackTrace();
			Gdx.app.debug("Visualizer reflection", "Failed attempt at retrieving tempBuffer field.");
			Gdx.app.exit();
		}
	}
	
	/* (non-Javadoc)
	 * @see zero1hd.rhythmbullet.desktop.audio.visualizer.Visualizer#calcPCMData()
	 */
	@Override
	public void calcPCMData() {
		short chanVal;
		int bufferPosOffset = 0;
		if (mm.playbackIndexUpdate() != readWindowIndex) {
			bufferPosOffset = updateBufferPosition();
			if (bufferPosOffset < 0 && playingBuffer.limit()+bufferPosOffset > 0) {
				playingBuffer.position(playingBuffer.limit()+bufferPosOffset);
			}
		}
		
		for (int sid = 0; sid < getAudioPCM().length && sid < buffer.remaining(); sid++) {
			for (int channel = 0; channel < mm.getChannelCount(); channel ++) {
				if (playingBuffer.position() < playingBuffer.limit()) {
					if (getAudioPCM()[sid] < (chanVal = playingBuffer.get())) {
						getAudioPCM()[sid] = chanVal;
					}
				} else {
					if (getAudioPCM()[sid] < (chanVal = buffer.get())) {
						getAudioPCM()[sid] = chanVal;
					}
				}
			}
			getAudioPCM()[sid] /= Short.MAX_VALUE+1f;
		}
		readWindowIndex++;
		
		
		//Take down original buffer position so we don't need to sync again after...
		int originalPos = buffer.position();
		
		//Begin comparison
		buffer.rewind();
		if (compareBuffer.compareTo(buffer) != 0) {
			bufferChanged();
			
			
			//Begin copying current buffer to the comparison buffer
			compareBuffer.clear();
			compareBuffer.put(buffer);
			compareBuffer.flip();
		}
		
		//Reset buffer to proper position.
		buffer.position(originalPos);
	}
	
	private void bufferChanged() {
		playingBuffer.clear();
		playingBuffer.put(compareBuffer);
	}
	
	private int updateBufferPosition() {
		int pos = (int) ((alGetSourcef(sourceID, AL11.AL_SAMPLE_OFFSET))-buffer.limit()-mm.getChannelCount()*mm.getReadWindowSize());
		readWindowIndex = mm.getPlaybackIndexPosition()-1;
		if (pos < 0) {
			buffer.position(0);
		} else {
			buffer.position(pos);
		}
		return pos;
	}
	
	/* (non-Javadoc)
	 * @see zero1hd.rhythmbullet.desktop.audio.visualizer.Visualizer#setMM(zero1hd.rhythmbullet.audio.MusicManager)
	 */
	@Override
	public void setMM(MusicManager mm) {
		try {
			Field sourceIDField = ClassReflection.getDeclaredField(OpenALMusic.class, "sourceID");
			sourceIDField.setAccessible(true);
			sourceID = (int) sourceIDField.get(mm.getMusic());
		} catch (ReflectionException e) {
			e.printStackTrace();
		}
		int originalPos = buffer.position();

		playingBuffer = ShortBuffer.allocate(buffer.capacity());
		buffer.rewind();
		playingBuffer.put(buffer);
		playingBuffer.flip();
		
		compareBuffer = ShortBuffer.allocate(buffer.capacity());
		buffer.rewind();
		compareBuffer.put(buffer);
		compareBuffer.flip();
		
		buffer.position(originalPos);
		this.mm = mm;
		
		super.setMM(mm);
	}
	
	@Override
	public MusicManager getMM() {
		return mm;
	}
	
	@Override
	public void render(Batch batch, float parentAlpha) {
	}
	
	@Override
	public float[] getAudioPCMData() {
		return getAudioPCM();
	}
	
	@Override
	public void fft() {
		calculate();
	}
}
